@@ -44,6 +44,8 @@ CONTACTS_BY_COMPANY = "contacts_by_company"
 
 DEFAULT_CHUNK_SIZE = 1000 * 60 * 60 * 24
 
+V3_PREFIXES = {'hs_date_entered', 'hs_date_exited', 'hs_time_in'}
+
 CONFIG = {
     "access_token": None,
     "token_expires": None,
@@ -76,8 +78,10 @@ ENDPOINTS = {
     "deals_all":            "/deals/v1/deal/paged",
     "deals_recent":         "/deals/v1/deal/recent/modified",
     "deals_detail":         "/deals/v1/deal/{deal_id}",
-    "deals_v3_search":      "/crm/v3/objects/deals/search",
+
+    "deals_v3_batch_read":  "/crm/v3/objects/deals/batch/read",
     "deals_v3_properties":  "/crm/v3/properties/deals",
+
     "deal_pipelines":       "/deals/v1/pipelines",
 
     "campaigns_all":        "/email/public/v1/campaigns/by-id",
@@ -159,13 +163,10 @@ def get_field_schema(field_type, extras=False):
                 "value": get_field_type_schema(field_type),
             }
         }
-def parse_custom_schema(entity_name, data, force_extras=None):
-    if force_extras is not None:
-        extras = force_extras
-    else:
-        extras = entity_name != 'contacts'
+
+def parse_custom_schema(entity_name, data):
     return {
-        field['name']: get_field_schema(field['type'], extras)
+        field['name']: get_field_schema(field['type'], entity_name != 'contacts')
         for field in data
     }
 
@@ -175,8 +176,7 @@ def get_custom_schema(entity_name):
 
 def get_v3_schema(entity_name):
     url = get_url("deals_v3_properties")
-    return parse_custom_schema(entity_name, request(url).json()['results'], force_extras=False)
-
+    return parse_custom_schema(entity_name, request(url).json()['results'])
 
 def get_abs_path(path):
     return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)
@@ -201,7 +201,7 @@ def load_schema(entity_name):
         if entity_name in ["deals"]:
             v3_schema = get_v3_schema(entity_name)
             for key, value in v3_schema.items():
-                if 'hs_date_entered' in key or 'hs_date_exited' in key:
+                if any(prefix in key for prefix in V3_PREFIXES):
                     custom_schema[key] = value
 
         # Move properties to top level
@@ -265,17 +265,13 @@ def parse_source_from_url(url):
         return match.group(1)
     return None
 
-
-@backoff.on_exception(backoff.constant,
-                      (requests.exceptions.RequestException,
-                       requests.exceptions.HTTPError),
-                      max_tries=5,
-                      jitter=None,
-                      giveup=giveup,
-                      on_giveup=on_giveup,
-                      interval=10)
-def request(url, params=None):
-
+def get_params_and_headers(params):
+    """
+    This function makes a params object and headers object based on the
+    authentication values available. If there is an `hapikey` in the config, we
+    need that in `params` and not in the `headers`. Otherwise, we need to get an
+    `access_token` to put in the `headers` and not in the `params`
+    """
     params = params or {}
     hapikey = CONFIG['hapikey']
     if hapikey is None:
@@ -289,6 +285,21 @@ def request(url, params=None):
     if 'user_agent' in CONFIG:
         headers['User-Agent'] = CONFIG['user_agent']
 
+    return params, headers
+
+
+@backoff.on_exception(backoff.constant,
+                      (requests.exceptions.RequestException,
+                       requests.exceptions.HTTPError),
+                      max_tries=5,
+                      jitter=None,
+                      giveup=giveup,
+                      on_giveup=on_giveup,
+                      interval=10)
+def request(url, params=None):
+
+    params, headers = get_params_and_headers(params)
+
     req = requests.Request('GET', url, params=params, headers=headers).prepare()
     LOGGER.info("GET %s", req.url)
     with metrics.http_request_timer(parse_source_from_url(url)) as timer:
@@ -321,19 +332,8 @@ def lift_properties_and_versions(record):
     return record
 
 def post_search_endpoint(url, data, params=None):
-    params = params or {}
-    hapikey = CONFIG['hapikey']
-    if hapikey is None:
-        if CONFIG['token_expires'] is None or CONFIG['token_expires'] < datetime.datetime.utcnow():
-            acquire_access_token_from_refresh_token()
-        headers = {'Authorization': 'Bearer {}'.format(CONFIG['access_token'])}
-    else:
-        params['hapikey'] = hapikey
-        headers = {}
-
-    if 'user_agent' in CONFIG:
-        headers['User-Agent'] = CONFIG['user_agent']
 
+    params, headers = get_params_and_headers(params)
     headers['content-type'] = "application/json"
 
     with metrics.http_request_timer(url) as timer:
@@ -348,6 +348,40 @@ def post_search_endpoint(url, data, params=None):
 
     return resp
 
+def merge_responses(v1_data, v3_data):
+    for v1_record in v1_data:
+        v1_id = v1_record.get('dealId')
+        for v3_record in v3_data:
+            v3_id = v3_record.get('id')
+            if str(v1_id) == v3_id:
+                v1_record['properties'] = {**v1_record['properties'],
+                                           **v3_record['properties']}
+
+def process_v3_deals_records(v3_data):
+    """
+    This function:
+    1. filters out fields that don't contain 'hs_date_entered_*' and
+       'hs_date_exited_*'
+    2. changes a key value pair in `properties` to a key paired to an
+       object with a key 'value' and the original value
+    """
+    transformed_v3_data = []
+    for record in v3_data:
+        new_properties = {field_name : {'value': field_value}
+                          for field_name, field_value in record['properties'].items()
+                          if any(prefix in field_name for prefix in V3_PREFIXES)}
+        transformed_v3_data.append({**record, 'properties' : new_properties})
+    return transformed_v3_data
+
+def get_v3_deals(v3_fields, v1_data):
+    v1_ids = [{'id': str(record['dealId'])} for record in v1_data]
+
+    v3_body = {'inputs': v1_ids,
+               'properties': v3_fields,}
+    v3_url = get_url('deals_v3_batch_read')
+    v3_resp = post_search_endpoint(v3_url, v3_body)
+    return v3_resp.json()['results']
+
 #pylint: disable=line-too-long
 def gen_request(STATE, tap_stream_id, url, params, path, more_key, offset_keys, offset_targets, v3_fields=None):
     if len(offset_keys) != len(offset_targets):
@@ -360,18 +394,16 @@ def gen_request(STATE, tap_stream_id, url, params, path, more_key, offset_keys,
         while True:
             data = request(url, params).json()
 
+            if data.get(path) is None:
+                raise RuntimeError("Unexpected API response: {} not in {}".format(path, data.keys()))
+
             if v3_fields:
-                url = get_url('deals_v3_search')
-                body = {"properties": v3_fields, "limit": 100}
-                v3_data = post_search_endpoint(url, body)
-                additional_fields = {}
-                for item in v3_data.json()['results']:
-                    # We nest `y` under 'value' in order to match the
-                    # schema and the shape of other fields in 'properties'
-                    additional_fields[int(item['id'])] = {key:{'value': value} for key,value in item['properties'].items()
-                                                          if ('hs_date_entered' in key or 'hs_date_exited' in key)}
-                for item in data[path]:
-                    item['properties'] = {**item['properties'], **additional_fields.get(item['dealId'])}
+                v3_data = get_v3_deals(v3_fields, data[path])
+
+                # The shape of v3_data is different than the V1 response,
+                # so we transform v3 to look like v1
+                transformed_v3_data = process_v3_deals_records(v3_data)
+                merge_responses(data[path], transformed_v3_data)
 
             for row in data[path]:
                 counter.increment()
@@ -469,6 +501,10 @@ def _sync_contacts_by_company(STATE, ctx, company_id):
     with Transformer(UNIX_MILLISECONDS_INTEGER_DATETIME_PARSING) as bumble_bee:
         with metrics.record_counter(CONTACTS_BY_COMPANY) as counter:
             data = request(url, default_contacts_by_company_params).json()
+
+            if data.get(path) is None:
+                raise RuntimeError("Unexpected API response: {} not in {}".format(path, data.keys()))
+
             for row in data[path]:
                 counter.increment()
                 record = {'company-id' : company_id,
@@ -553,7 +589,7 @@ def sync_deals(STATE, ctx):
     max_bk_value = start
     LOGGER.info("sync_deals from %s", start)
     most_recent_modified_time = start
-    params = {'count': 250,
+    params = {'limit': 100,
               'includeAssociations': False,
               'properties' : []}
 
@@ -579,9 +615,11 @@ def sync_deals(STATE, ctx):
         params['allPropertiesFetchMode'] = 'latest_version'
 
         # Grab selected `hs_date_entered/exited` fields to call the v3 endpoint with
-        v3_fields = [x[1].replace('property_', '')
-                     for x,y in mdata.items() if x and (y.get('selected') == True or has_selected_properties)
-                     and ('hs_date_entered' in x[1] or 'hs_date_exited' in x[1])]
+        v3_fields = [breadcrumb[1].replace('property_', '')
+                     for breadcrumb, mdata_map in mdata.items()
+                     if breadcrumb
+                     and (mdata_map.get('selected') == True or has_selected_properties)
+                     and any(prefix in breadcrumb[1] for prefix in V3_PREFIXES)]
 
     url = get_url('deals_all')
     with Transformer(UNIX_MILLISECONDS_INTEGER_DATETIME_PARSING) as bumble_bee:
@@ -665,6 +703,9 @@ def sync_entity_chunked(STATE, catalog, entity_name, key_properties, path):
                     data = request(url, params).json()
                     time_extracted = utils.now()
 
+                    if data.get(path) is None:
+                        raise RuntimeError("Unexpected API response: {} not in {}".format(path, data.keys()))
+
                     for row in data[path]:
                         counter.increment()
                         record = bumble_bee.transform(lift_properties_and_versions(row), schema, mdata)
@@ -1002,6 +1043,7 @@ def load_discovered_schema(stream):
     # The engagements stream has nested data that we synthesize; The engagement field needs to be automatic
     if stream.tap_stream_id == "engagements":
         mdata = metadata.write(mdata, ('properties', 'engagement'), 'inclusion', 'automatic')
+        mdata = metadata.write(mdata, ('properties', 'lastUpdated'), 'inclusion', 'automatic')
 
     return schema, metadata.to_list(mdata)
 
